/*
 * This program starts an other program with the same basename
 * as the starter but with the extension .bin.
 * 
 * Before the child program is executed, BEGINLIBPATH is set to
 * the directory the starter program has been executed from. An
 * existing BEGINLIBPATH is preserved, the new path is added in
 * front of the existing paths.
 * 
 * Once BEGINLIBPATH is set, LIBPATHTRICT=T will be activated.
 * By using the command line option -nolibpathstrict, this step
 * can be skipped.
 * 
 * Usage: STARTER.EXE [-?] [-nolibpathstrict] [arguments]
 *        -?               : Shows this info
 *        -nolibpathstrict : Don't use LIBPATHSTRICT
 *        arguments        : Arguments passed to the process
 * 
 * The starter program can have any name. It is important that
 * the program to be started must have the same basename as the
 * starter.
 * 
 * License: MOZILLA PUBLIC LICENSE Version 1.1
 * Author : Heiko Nitzsche
 * Date   : 03-Feb-2007
 * Version: 1.0 beta1
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define INCL_DOSERRORS
#define INCL_DOSMISC
#define INCL_DOSMODULEMGR
#define INCL_DOSPROCESS
#define INCL_WINDIALOGS
#include <os2.h>

/********************************************/

static int  IS_PM_APP = 0;

static int  libpathStrictActive        = 0;
static int  beginlibpathActive         = 0;
static char initialBeginLibpath [1025] = { 0 };
static char initialLibpathStrict[1025] = { 0 };

/********************************************/

static char * stringdup(const char * s)
{
    /* we could use strdup() here but this is an extension, not in ANSI */
    char * rs = NULL;
    if (s != NULL)
    {
        rs = (char *) malloc(strlen(s) + 1);
        if (rs != NULL)
        {
            strcpy(rs, s);
        }
    }
    return rs;
}

static char * string_append(char * target, const char * source)
{
    char * rs = target;
    if ((rs != NULL) && (source != NULL))
    {
        rs = (char *) malloc(strlen(target) + strlen(source) + 1);
        if (rs != NULL)
        {
            strcpy(rs, target);
            strcat(rs, source);
        }
        free(target); target = NULL;
    }
    return rs;
}

/********************************************/

static void showMessage(const char * title, const char * message, int asError)
{
    if (IS_PM_APP)
    {
        /* show message box with usage */
        const HAB hab = WinInitialize(0);
        if (hab != 0)
        {
            const HMQ hmq = WinCreateMsgQueue(hab, 0);
            if (hmq != 0)
            {
                WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, (PSZ)message, (PSZ)title, 0, 
                              MB_MOVEABLE | MB_OK | (asError ? MB_ERROR : MB_NOICON));
                WinDestroyMsgQueue(hmq);
            }
            WinTerminate(hab);
        }
    }
    else
    {
        fprintf(asError ? stderr : stdout, "%s\n", message);
    }
}

static void showErrorMessage(const char * message)
{
    showMessage("Error", message, 1);
}

/********************************************/

static void showUsage(const char * progname, const char * childProgname)
{
    char msg[CCHMAXPATH + 400]  = { 0 };
    const char * pProgname      = strrchr(progname, '\\');
    const char * pChildProgname = strrchr(childProgname, '\\');

    if ((pProgname == NULL) || (pChildProgname == NULL))
    {
        showErrorMessage("Out of memory.");
        return;
    }
    pProgname++;
    pChildProgname++;

    if (IS_PM_APP)
    {
        /* show message box with usage */
        sprintf(msg, "'%s' is a launcher program for '%s'."                              \
                     "It activates LIBPATHSTRICT before it starts the child process "    \
                     "and thus enforces that DLLs are loaded from the directory where "  \
                     "'%s' is located.\n\n"                                              \
                     "%s [-?] [-nolibpathstrict] [arguments]\n"                          \
                     "-? : Shows this info\n"                                            \
                     "-nolibpathstrict : Don't use LIBPATHSTRICT\n"                      \
                     "arguments : Arguments passed to the process\n "                    \
                     , pProgname, pChildProgname, pChildProgname, pProgname);
    }
    else
    {
        sprintf(msg, "\n'%s' is a launcher program for '%s'.\n"                          \
                     "It activates LIBPATHSTRICT before it starts the child process\n"   \
                     "and thus enforces that DLLs are loaded from the directory where\n" \
                     "'%s' is located.\n\n"                                              \
                     "Usage: %s [-?] [-nolibpathstrict] [arguments]\n"                   \
                     "       -?               : Shows this info\n"                       \
                     "       -nolibpathstrict : Don't use LIBPATHSTRICT\n"               \
                     "       arguments        : Arguments passed to the process\n "      \
                     , pProgname, pChildProgname, pChildProgname, pProgname);
    }

    showMessage("Usage", msg, 0);
}

/********************************************/

static int saveAndActivateLibpaths(const char * fullProgramPathWithName,
                                   const int    enableLibpathStrict)
{
    APIRET rc;

    /* check that we don't accidently override our backup paths */
    if (libpathStrictActive || beginlibpathActive)
    {
        return 1;
    }

    beginlibpathActive  = 0;
    libpathStrictActive = 0;
    memset(initialBeginLibpath , 0, sizeof(initialBeginLibpath));
    memset(initialLibpathStrict, 0, sizeof(initialLibpathStrict));

    /* remember BEGINLIBPATH */
    rc = DosQueryExtLIBPATH((PSZ) initialBeginLibpath, BEGIN_LIBPATH);
    if (rc == 0)
    {
        /* remember LIBPATHSTRICT */
        rc = DosQueryExtLIBPATH((PSZ) initialLibpathStrict, LIBPATHSTRICT);
        if (rc == 0)
        {
            char * extBeginlibpath = NULL;
            char * basename = stringdup(fullProgramPathWithName);
            if (basename == NULL)
            {
                return 1;
            }

            /* strip filename from path */
            {
                char drive[_MAX_DRIVE+1] = { 0 };
                char dir  [_MAX_DIR+1]   = { 0 };
            
                _splitpath(basename, drive, dir, NULL, NULL);
    
                /* if no path was found, set the current path as ".\." */
                if ((strlen(drive) == 0) || (strlen(dir) == 0))
                {
                    _makepath(basename, NULL, ".\\.", NULL, NULL);
                }
                else
                {
                    _makepath(basename, drive, dir, NULL, NULL);
                }
            }
        
            /* --------------------------------------------- */
        
            /* merge the executable path (basename) with the queried BEGINLIBPATH */
            extBeginlibpath = (char *) malloc(strlen(basename) + 1 + strlen(initialBeginLibpath) + 1);
            if (extBeginlibpath == NULL)
            {
                free(basename);
                return 1;
            }

            strcpy(extBeginlibpath, basename);

            free(basename); basename = NULL;

            if (strlen(initialBeginLibpath) > 0)
            {
                strcat(extBeginlibpath, ";");
                strcat(extBeginlibpath, initialBeginLibpath);
            }
        
            /* set BEGINLIBPATH */
            rc = DosSetExtLIBPATH((PSZ) extBeginlibpath, BEGIN_LIBPATH);
            free(extBeginlibpath); extBeginlibpath = NULL;
            if (rc == 0)
            {
                beginlibpathActive = 1;

                /* set LIBPATHSTRICT=T */
                if (enableLibpathStrict)
                {
                    rc = DosSetExtLIBPATH((PSZ) "T", LIBPATHSTRICT);
                    if (rc == 0)
                    {
                        libpathStrictActive = 1;
                    }
                }
            }
        }
    }
    return 0;
}

/********************************************/

static void restoreLibpaths(void)
{
    if (beginlibpathActive)
    {
        /* reset BEGINLIBPATH to initial settings */
        DosSetExtLIBPATH((PSZ) initialBeginLibpath, BEGIN_LIBPATH);
        beginlibpathActive = 0;
    }
    if (libpathStrictActive)
    {
        /* reset LIBPATHSTRICT to initial setting */
        DosSetExtLIBPATH((PSZ) initialLibpathStrict, LIBPATHSTRICT);
        libpathStrictActive = 0;
    }
}

/********************************************/

static APIRET runAsChild(const char  * fullProgramPathWithName,
                         const char  * args,
                               ULONG * returnCode)
{
    CHAR          loadError[CCHMAXPATH+1] = { 0 };
    RESULTCODES   childRc                 = { 0 };
    APIRET        rc;
    CHAR        * callArgs = NULL;
    
    if ((fullProgramPathWithName == NULL) || (fullProgramPathWithName == NULL))
    {
        return ERROR_INVALID_FUNCTION;
    }

    callArgs = (CHAR *) calloc(strlen(fullProgramPathWithName) + 1 + strlen(args) + 2 + 1,
                               sizeof(CHAR));
    if (callArgs == NULL)
    {
        return ERROR_NOT_ENOUGH_MEMORY;
    }
    strcpy(callArgs, fullProgramPathWithName);
    strcpy(callArgs + strlen(fullProgramPathWithName) + 1, args);

    rc = DosExecPgm(loadError,                      /* Object name buffer           */
                    sizeof(loadError),              /* Length of object name buffer */
                    EXEC_SYNC,                      /* synchronous                  */
                    (PSZ) callArgs,                 /* Argument string              */
                    NULL,                           /* Environment string           */
                    &childRc,                       /* Termination codes            */
                    (PSZ) fullProgramPathWithName); /* Program file name            */
    free(callArgs); callArgs = NULL;

    *returnCode = childRc.codeResult;
    return rc;
}

/********************************************/

static char * parseArgumentsAndCreateString(const int argc, const char * argv[],
                                            int * libpathstrictEnabled)    
{
    char * childArgv = stringdup("");

    *libpathstrictEnabled = 0;
    if ((childArgv != NULL) && (argv != NULL))
    {
        int i;

        /* Check whether the user has explicitly disabled LIBPATHSTRICT usage */
        *libpathstrictEnabled = 1;
    
        for (i = 1; i < argc; i++)
        {
            if ((strcmp(argv[i], "-nolibpathstrict") == 0) ||
                (strcmp(argv[i], "/nolibpathstrict") == 0))
            {
                /* skip LIBPATHSTRICT activation */
                *libpathstrictEnabled = 0;

                /* the parameter will NOT be forwarded to the child process */
            }
            else
            {
                /* the parameter will be forwarded to the child process */
                char * s = string_append(childArgv, argv[i]);
                if (s == NULL)
                {
                    free(childArgv); childArgv = NULL;
                    return NULL;
                }
                childArgv = s;
    
                s = string_append(childArgv, " ");
                if (s == NULL)
                {
                    free(childArgv); childArgv = NULL;
                    return NULL;
                }
                childArgv = s;
            }
        }
    }
    return childArgv;
}

/********************************************/
/********************************************/

int main(const int argc, const char * argv[])
{
    char * childProcessName      = NULL;
    char * childArgv             = NULL;
    char   exePath[CCHMAXPATH+1] = { 0 };
    int    enableLibpathstrict   = 0;
    ULONG  childReturnCode       = 0;
            
    {
        /* GCC runtime does not provide the fully qualified name. */
        /* But we can query it from the OS. */
        PTIB ptib;
        PPIB ppib;
        DosGetInfoBlocks( &ptib, &ppib);

        /* remember whether we don't have a console but rather the PM */
        IS_PM_APP = (ppib->pib_ultype == 3);
    
        if (DosQueryModuleName(ppib->pib_hmte, sizeof(exePath), exePath) ||
            (strrchr(exePath, '\\') == 0))
        {
            showErrorMessage("Cannot determine process filename.");
            return 1;
        }
        childProcessName = stringdup(exePath);
    }

    /* -------------------------------------- */

    /* build the executable name of the child */
    if (childProcessName == NULL)
    {
        showErrorMessage("Out of memory.");
        return 1;
    }
    /* patch extension */
    {
        char * pChildProcessName = strrchr(childProcessName, '.');
        if (pChildProcessName == NULL)
        {
            free(childProcessName); childProcessName = NULL;
            showErrorMessage("Illegal process name.");
            return 1;
        }
        if (strcmp(pChildProcessName, ".EXE") != 0)
        {
            free(childProcessName); childProcessName = NULL;
            showErrorMessage("Illegal process name.");
            return 1;
        }
        /* replace ".EXE" by ".BIN" keeping the original case */
        pChildProcessName[1] = 'B';
        pChildProcessName[2] = 'I';
        pChildProcessName[3] = 'N';
    }

    /* -------------------------------------- */

    /* check if usage was requested */
    if (argc == 2)
    {
        if ((strcmp(argv[1], "-?") == 0) ||
            (strcmp(argv[1], "/?") == 0))
        {
            showUsage(exePath, childProcessName);
            free(childProcessName); childProcessName = NULL;
            return 0;
        }
    }

    /* -------------------------------------- */

    childArgv = parseArgumentsAndCreateString(argc, argv,
                                              &enableLibpathstrict);
    if (childArgv == NULL)
    {
        free(childProcessName); childProcessName = NULL;
        showErrorMessage("Out of memory.");
        return 1;
    }
        
    /* -------------------------------------- */

    /* Start the process as synchronous child to simulate the original process behaviour */
    {
        APIRET runRc = 0;
    
        /* Activate LIBPATHSTRICT */
        if (saveAndActivateLibpaths(childProcessName, enableLibpathstrict))
        {
            if (enableLibpathstrict && (! libpathStrictActive))
            {
                showErrorMessage("LIBPATHSTRICT could not be activated.");
            }
            else
            {
                showErrorMessage("BEGINLIBPATH could not be set.");
            }
            free(childProcessName); childProcessName = NULL;
            free(childArgv); childArgv = NULL;
            return 1;
        }
        /* Start the process */
        runRc = runAsChild(childProcessName, childArgv, &childReturnCode);

        free(childArgv); childArgv = NULL;

        /* restore initial settings */
        restoreLibpaths();
    
        switch(runRc)
        {
            case NO_ERROR:
                break;
                
            case ERROR_FILE_NOT_FOUND:
                {
                    char msg[CCHMAXPATH+30] = { 0 };
                    sprintf(msg, "Could not find file \"%s\".", childProcessName);
                    showErrorMessage(msg);
                }
                free(childProcessName); childProcessName = NULL;
                return runRc;
    
            default:
                {
                    char msg[CCHMAXPATH+40] = { 0 };
                    sprintf(msg, "Could not execute \"%s\" : RC=%lu", childProcessName, runRc);
                    showErrorMessage(msg);
                }
                free(childProcessName); childProcessName = NULL;
                return runRc;
        }
    }

    free(childProcessName); childProcessName = NULL;

    return childReturnCode;
}


